import struct

class ValueToLargeError(BaseException):
  pass

class ReadOnlyProperty(BaseException):
  pass

class RegBase:
  def __init__(self, source):
    self.children = {}
    self.source = source
    self.unit = ''
  def register(self, name, child):
    self.children[name] = child
  def mapv(self, obj, value):
    return value
  def splitr(self, obj, value, soFar={}):
    mv = self.mapv(obj, value)
    res = soFar | {self.name: mv}
#    print(f'{self.name}.split in:{soFar}',)
    return self.splitrc(obj, mv, res)
  def splitrc(self, obj, value, res={}):
    if len(self.children) > 0:
      for c in self.children:
        if self.children[c] is not None:
          res |= self.children[c].splitr(obj, value, res)
#    print(f'{self.name}.splitc res {res}')
    return res
  def __set_name__(self, owner, name):
    self.owner = owner
    self.name = name
    self.source.register(name, self)
  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    return self.mapv(obj, self.source.__get__(obj))
  def __str__(self):
    return self.tstr()

class RegDescriptor(RegBase):
  def __init__(self, source):
    super().__init__(source)
    self.bank = 0
    self.reg = source
    self.mask = 0xFF
    self.width = 8
  def __set_name__(self, owner, name):
    self.name = name
  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    return obj._get(self.source)
  def __set__(self, obj, value):
    return obj._set(self.source, value)
  def tstr(self):
    return f'{self.name}(0x{self.source:X})'
  def __str__(self):
    return self.tstr()

class RegMBDescriptor(RegBase):
  def __init__(self, source, length):
    super().__init__(source)
    self.length = length
    self.bank = 0
  def __set_name__(self, owner, name):
    self.name = name
    self._owner = owner
  def __get__(self, obj, objtype=None):
#    print(f'RegMB.__get__, {self.source:X}, {self.length}, {repr(obj)}, {self._owner}')
    if obj is None:
      return self
    return obj._spi.getbytes(self.source, self.length)
  def tstr(self):
    return f'{self.name}(0x{self.source:X}..0x{self.source+self.length:X})'
  def __str__(self):
    return self.tstr()

class Reg3DDataDescriptor(RegBase):
  def __init__(self, source, endn='>'):
    super().__init__(source)
    self.length = 6
    self.bank = 0
    self.endn = endn
  def __set_name__(self, owner, name):
    self.name = name
  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    return obj._spi.get3x2(self.source, self.endn)
  def tstr(self):
    return f'{self.name}(0x{self.source:X}..0x{self.source+self.length:X})'
  def __str__(self):
    return self.tstr()

class ArrayItemDescriptor(RegBase):
  def __init__(self, source, index):
    super().__init__(source)
    self.index = index
  def mapv(self, obj, value):
    return value[self.index]
  def tstr(self):
    return f'{self.source.tstr()}[{self.index}]'

class DataDescriptor(RegBase):
  def __init__(self, length):
    super().__init__(None)
    self.length = length
  def __get__(self, obj, objtype=None):
    return obj.getData()

class DataSubsetDescriptor8B(RegBase):
  width = 8
  def __init__(self, source, offs):
    super().__init__(source)
    self.length = 1
    self.offs = offs
    self.bank = source.bank
  def mapv(self, obj, value):
    #    print(f'{self.name} get byte {self.offs} of {len(value)} B')
    return value[self.offs]
  def tstr(self):
    return f'{self.source.tstr()}[{self.offs}]'

class DataSubsetDescriptor16B(RegBase):
  width = 16
  def __init__(self, source, offs, endn='>', dtype='h'):
    super().__init__(source)
    self.length = 2
    self.offs = offs
    self.endn = endn
    self.dtype = dtype
    self.bank = source.bank
  def mapv(self, obj, value):
    return struct.unpack(self.endn + self.dtype, struct.pack('BB', *value[self.offs:(self.offs+self.length)]))[0]
  def tstr(self):
    return f'{self.source.tstr()}[{self.offs}:{self.offs+self.length}]'

class RegInt16Descriptor(RegMBDescriptor):
  def __init__(self, source, end='>', dtype='h'):
    super().__init__(source, 2)
    self.end = end
    self.dtype = dtype
  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    return obj._spi.readRegInt16(self.source, end=self.end, dtype=self.dtype)
  def __set__(self, obj, value):
    return obj._spi.writeRegInt16(self.source, value, end=self.end, dtype=self.dtype)

class RegBankDescriptor(RegDescriptor):
  def __init__(self, bank, source):
    super().__init__(source)
    self.bank = bank
  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    return obj._getb(self.bank, self.source)
  def __set__(self, obj, value):
    return obj._setb(self.bank, self.source, value)
  def tstr(self):
    return f'{self.name}({self.bank}:0x{self.source:X})'

class RegBankMBDescriptor(RegBankDescriptor):
  def __init__(self, bank, source, length):
    super().__init__(bank, source)
    self.length = length

class RegBankInt24Descriptor(RegBankMBDescriptor):
  def __init__(self, bank, source):
    super().__init__(bank, source, 3)
  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    with obj._setUserBank(self.bank):
      return obj._spi.readRegInt24(self.source)
  def __set__(self, obj, value):
    with obj._setUserBank(self.bank):
      return obj._spi.writeRegInt24(self.source)
  def tstr(self):
    return f'{self.name}({self.bank}:0x{self.source:X}..0x{self.source+self.length:X})'
  def __str__(self):
    return self.tstr()

class BitfieldDescriptor(RegBase):
  def __init__(self, source, width, offs=0):
    super().__init__(source)
    self.offs = offs
    self.width = width
    self.mask = (1 << self.width) - 1
    self.bank = source.bank
  def mapv(self, obj, value):
    return (value >> self.offs) & self.mask
  def __set__(self, obj, value):
    cur = self.source.__get__(obj)
    cur &= ~(self.mask << self.offs)
    cur |= (int(value) & self.mask) << self.offs
    self.source.__set__(obj, cur)
  def tstr(self):
    return str(self.source) + '[' + str(self.offs) + ':' + str(self.offs+self.width) + ']'
  def __str__(self):
    return self.name + ': ' + self.tstr()

class JointBitfieldDescriptor(RegBase):
  def __init__(self, sources):
    super().__init__(None)
    self.sources = sources
    self.widths = [ s.width for s in self.sources ]
    self.width = sum(self.widths)
    self.mask = (1 << self.width) - 1
    self.bank = sources[0].bank
  def __set_name__(self, owner, name):
    self.name = name
    for s in self.sources[0:-1]:
      s.register(name, None)
    self.sources[-1].register(name, self)
    self.snames = [ s.name for s in self.sources ]
  def mapv(self, obj, values):
    val = 0
    shift = 0
    for (i, w) in enumerate(self.widths):
      val |= values[i] << shift
      shift += w
    return val
  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    vals = [s.__get__(obj) for s in self.sources]
    val = self.mapv(obj, vals)
    return val
  def __set__(self, obj, value):
    if (self.mask & value) != value:
      raise(ValueToLargeError(f'int value {value} is too large for joint int of {self.width} bits'))
    for s in self.sources:
      part = value & s.mask
      #print(f'jfd {s} mask={s.mask} value={value} part={part}')
      s.__set__(obj, part)
      value >>= s.width
  def splitr(self, obj, value, soFar):
    #    mv = self.mapv(obj, values)
#    print(f'\nJBD {self.name}.split in:{soFar} (sources: {self.snames}, {self.widths})')
    mv = self.mapv(obj, [soFar[v] for v in self.snames ])
    res = soFar | {self.name: mv}
#    print(f'\nJBD {self.name}.split interm.: {res})')
    res = super().splitrc(obj, mv, res)
#    print(f'JBD {self.name}.split res {res}\n')
    return res
  def tstr(self):
    return '|'.join([s.tstr() for s in self.sources])
  def __str__(self):
    return self.name + ': ' + self.tstr()

class SignedIntDescriptor(RegBase):
  def __init__(self, source):
    super().__init__(source)
    self.width = self.source.width
    self.mask = (1 << self.source.width) - 1
    self.sign = 1 << (self.source.width - 1)
    self.ceil = 1 << self.source.width
  def _getval(self, val):
    if val & self.sign:
      val = val - self.ceil
    return val
  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    val = self.source.__get__(obj)
    return self._getval(val)
  def __set__(self, obj, value):
    if self._getval(self.mask & value) != value:
      raise(ValueToLargeError(f'int value {value} ({self.mask & value}) is too large for signed int with width of {self.width} bits'))
    # print(f'SignedInt({self.width} bit).set value={value} => {self.mask & value}')
    self.source.__set__(obj, self.mask & value)
  def tstr(self):
    return f'Signed({self.source.tstr()})'
  def __str__(self):
    return self.name + ': ' + self.tstr()

class AliasDescriptor(RegBase):
  def __init__(self, source):
    super().__init__(source)
    self.unit = source.unit
  def mapv(self, obj, value):
    return value
  def __set__(self, obj, value):
    self.source.__set__(obj, value)
  def tstr(self):
    return self.source.tstr()
  def __str__(self):
    return self.name + ': ' + self.tstr()

class BitDescriptor(RegBase):
  def __init__(self, source, offs):
    super().__init__(source)
    self.offs = offs
    self.mask = (1 << offs)
    self.bank = source.bank
  def mapv(self, obj, value):
    return (value & self.mask) != 0
  def __set__(self, obj, value):
    value = 1 if value else 0
    cur = self.source.__get__(obj)
    cur &= ~(1 << self.offs)
    cur |= (value << self.offs)
    self.source.__set__(obj, cur)
  def tstr(self):
    return self.source.tstr() + '[' + str(self.offs) + ']'
  def __str__(self):
    return self.name + ': ' + self.tstr()

class NamedDescriptor(RegBase):
  def __init__(self, source, names):
    super().__init__(source)
    self.names = names
  def mapv(self, obj, value):
    return self.names[value] if value in self.names else None
  def __set__(self, obj, value):
    if type(value) == type(''):
      value = list(self.names.values()).index(value)
    self.source.__set__(obj, value)
  def tstr(self):
    return self.source.tstr() + ' = ' + str(list(self.names.values()))
  def __str__(self):
    return self.name + ': ' + self.tstr()

class MappedDescriptor(NamedDescriptor):
  def __init__(self, source, names, unit):
    super().__init__(source, names)
    self.unit = unit
    lv = list(names.values())
    lk = list(names.keys())
    self.invmap = { lv[i]: lk[i] for i in range(len(lk)) }
  def __set__(self, obj, value):
    if obj is None:
      return self
    valint = self.invmap[value]
    self.source.__set__(obj, valint)

class DiscreteDescriptor(RegBase):
  def __init__(self, source, values, unit):
    super().__init__(source)
    self.values = [v*1.0 if v is not None else None for v in values]
    self.unit = unit
  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    return self.values[self.source.__get__(obj)]
  def findClosest(self, vals, v):
    diffPairs = [(1e120 if w is None else abs(w - v),i) for (i,w) in enumerate(vals)]
    minPair = min(diffPairs, key=lambda x: x[0])
    return minPair
  def __set__(self, obj, value):
    (best, ind) = self.findClosest(self.values, value)
    self.source.__set__(obj, ind)
  def tstr(self):
    return self.source.tstr() + ' = ' + str(self.values)
  def __str__(self):
    return self.name + ': ' + self.tstr()

class CalculatedDescriptor(RegBase):
  def __init__(self, source, getFun, unit):
    super().__init__(source)
    self.getFun = getFun
    self.unit = unit
  def mapv(self, obj, value):
    return self.getFun(obj, value)
  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    return self.mapv(obj, self.source.__get__(obj))
  def __set__(self, obj, value):
    raise ReadOnlyProperty('Cannot set calculated property')
  def tstr(self):
    return self.source.tstr() + ' = ' + str(self.getFun) + '(x)'
  def __str__(self):
    return self.name + ': ' + self.tstr()

class ConversionDescriptor(CalculatedDescriptor):
  def __init__(self, source, getFun, setFun, unit):
    super().__init__(source, getFun, unit)
    self.setFun = setFun
  def __set__(self, obj, value):
    if obj is None:
      return self
    return self.source.__set__(obj, self.setFun(obj, value))
  def tstr(self):
    return self.source.tstr() + ' = ' + str(self.getFun) + '(x)'
  def __str__(self):
    return self.name + ': ' + self.tstr()

class ScaledDescriptor(ConversionDescriptor):
  def __init__(self, source, scale, unit):
    self.scale = scale
    getFun = lambda _, x: x * scale
    setFun = lambda _, x: int(round(x / scale))
    super().__init__(source, getFun, setFun, unit)
  def tstr(self):
    return self.source.tstr() + f' => f(x) = x * {self.scale} {self.unit}'
  def __str__(self):
    return self.name + ': ' + self.tstr()

class ArrayNDescriptor(RegBase):
  def __init__(self, sources):
    super().__init__(None)
    self.sources = sources
    sunits = sources[0].unit
    if all( [ s == sunits[0] for s in sunits ] ):
      self.unit = sources[0].unit
    else:
      self.unit = ''
  def __set_name__(self, owner, name):
    self.name = name
    for s in self.sources[0:-1]:
      s.register(name, None)
    self.sources[-1].register(name, self)
  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    vals = [s.__get__(obj) for s in self.sources]
    return vals
  def __set__(self, obj, values):
    if type(values) == type([]) or type(values) == type(tuple()):
      assert(len(values) == len(self.sources))
      for i in range(len(self.sources)):
        self.sources[i].__set__(obj, values[i])
    else:
      for i in range(len(self.sources)):
        self.sources[i].__set__(obj, values)
  def tstr(self):
    indent = ' ' *(len(self.name) + 3)
    return '[' + f',\n{indent}'.join([s.tstr() for s in self.sources]) + ']'
  def __str__(self):
    return self.name + ': ' + self.tstr()

class Array3Descriptor(ArrayNDescriptor):
  def __init__(self, sources):
    super().__init__(sources)
    assert(len(sources) == 3)

class NamedFieldDescriptor(RegBase):
  def __init__(self, source, names):
    super().__init__(source)
    self.unit = source.unit
    self.names = names
    assert(len(names) == len(self.source.sources))
  def __get__(self, obj, objtype=None):
    if obj is None:
      return self
    vals = self.source.__get__(obj)
    vals = { self.names[i]: vals[i] for i in range(len(self.names)) }
    return vals
  def tstr(self):
    descrDict = { self.names[i]: self.source.sources[i].tstr() for i in range(len(self.names)) }
    return str(descrDict)
  def __str__(self):
    return self.name + ': ' + self.tstr()


class FifoPkgFmt1Descriptor(RegBase):
  length = 8

  maskNoMsg = 0x80
  maskAccel = 0x40
  maskGyro  = 0x20
  mask20b   = 0x10

  maskContF1= maskAccel

  def __init__(self, basereg):
    super().__init__(basereg)
    self._fifoDataProp = getattr(FifoPkgFmt1Descriptor, 'FIFO_DATA')

  def __set_name__(self, owner, name):
    self._owner = owner
    self._name = name

  def __get__(self, obj, objtype=None):
#    print('FPkg1.__get__', obj, self._owner)
    if obj is None:
      return self
    self._spi = obj._spi
    blob = self.FIFO_DATA
    return self.parseData(obj, blob)

  def parseData(self, obj, blob):
    header = blob[0]
    if header & self.maskContF1 != self.maskContF1:
      print('Invalid FIFO pkt')
      return None

    accelInts      = struct.unpack('>' + 'h'*3, struct.pack('B'*6, *blob[1:7]))
    accel          = FifoPkgFmt1Descriptor.getAccelValues(obj, accelInts)
    temp           = FifoPkgFmt1Descriptor.getFifoTempValue(blob[7])

    data = {'values': accel + [ temp ] }
    return data

  def parseData2(self, obj, blob):
    data = self._fifoDataProp.splitr(obj, blob)
    if data['nomsg']:
      print('Invalid read: FIFO is empty')
      return None
    if not data['accel']:
      print('Configuration error: FIFO packet contains no accel data')
      return None
    if data['gyro']:
      print('Configuration error: FIFO packet contains gyro data')
      return None
    if data['is20b']:
      print('Configuration error: FIFO packet contains 20B data')
      return None
    data['values'] = [ data[v] for v in ['accelX', 'accelY', 'accelZ', 'temp' ] ]
    return data

  def getAccelValue(_obj, value):
    return _obj._getAccelValue(value)

  def getAccelValues(_obj, values):
    return _obj._getAccelValues(values)

  def getFifoTempValue(value):
    if value & 0x80:
      value = value - 0x100
    return value / 2.07 + 25

  def tstr(self):
    return 'Fifo Pkg Fmt I'

  FIFO_DATA      = RegMBDescriptor(0x30, length)
  header         = DataSubsetDescriptor8B(FIFO_DATA, 0)
  odrChGyro      = BitDescriptor(header, 0)
  odrChAccel     = BitDescriptor(header, 1)
  tmstFsync      = BitfieldDescriptor(header, 2, 2)
  is20b          = BitDescriptor(header, 4)
  gyro           = BitDescriptor(header, 5)
  accel          = BitDescriptor(header, 6)
  nomsg          = BitDescriptor(header, 7)

  accelXInt      = DataSubsetDescriptor16B(FIFO_DATA, 1)
  accelYInt      = DataSubsetDescriptor16B(FIFO_DATA, 3)
  accelZInt      = DataSubsetDescriptor16B(FIFO_DATA, 5)
  tempInt        = DataSubsetDescriptor8B(FIFO_DATA, 7)

  accelX         = CalculatedDescriptor(accelXInt, getAccelValue, 'g')
  accelY         = CalculatedDescriptor(accelYInt, getAccelValue, 'g')
  accelZ         = CalculatedDescriptor(accelZInt, getAccelValue, 'g')
  temp           = CalculatedDescriptor(tempInt, getFifoTempValue, '°C')

class FifoPkgFmt2Descriptor(FifoPkgFmt1Descriptor):

  length = 8

  def __init__(self, basereg):
    super().__init__(basereg)
    self._fifoDataProp = getattr(FifoPkgFmt2Descriptor, 'FIFO_DATA')

  maskContF2 = FifoPkgFmt1Descriptor.maskGyro

  def parseData(self, obj, blob):
    header = blob[0]
    if header & self.maskContF2 != self.maskContF2:
      print('Invalid FIFO fmt 2 pkt')
      return None

    gyroInts       = struct.unpack('>' + 'h'*3, struct.pack('B'*6, *blob[1:7]))

    gyroX          = FifoPkgFmt2Descriptor.getGyroValue(obj, gyroInts[0])
    gyroY          = FifoPkgFmt2Descriptor.getGyroValue(obj, gyroInts[1])
    gyroZ          = FifoPkgFmt2Descriptor.getGyroValue(obj, gyroInts[2])

    temp           = FifoPkgFmt1Descriptor.getFifoTempValue(blob[7])

    data = {'values': [ gyroX, gyroY, gyroZ, temp ] }
    return data

  def parseData2(self, obj, blob):
    data = self._fifoDataProp.splitr(obj, blob)
    if data['nomsg']:
      print('Invalid read: FIFO is empty')
      return None
    if data['accel']:
      print('Configuration error: FIFO packet contains accel data')
      return None
    if not data['gyro']:
      print('Configuration error: FIFO packet contains no gyro data')
      return None
    if data['is20b']:
      print('Configuration error: FIFO packet contains 20B data')
      return None
    data['values'] = [ data[v] for v in ['gyroX', 'gyroY', 'gyroZ', 'temp' ] ]
    return data

  def getGyroValue(_obj, value):
    return _obj._getGyroValue(value)

  def getGyroValues(_obj, values):
    return _obj._getGyroValues(values)

  def tstr(self):
    return 'Fifo Pkg Fmt II'

  FIFO_DATA      = RegMBDescriptor(0x30, length)
  header         = DataSubsetDescriptor8B(FIFO_DATA, 0)
  odrChGyro      = BitDescriptor(header, 0)
  odrChAccel     = BitDescriptor(header, 1)
  tmstFsync      = BitfieldDescriptor(header, 2, 2)
  is20b          = BitDescriptor(header, 4)
  gyro           = BitDescriptor(header, 5)
  accel          = BitDescriptor(header, 6)
  nomsg          = BitDescriptor(header, 7)

  gyroXInt       = DataSubsetDescriptor16B(FIFO_DATA, 1)
  gyroYInt       = DataSubsetDescriptor16B(FIFO_DATA, 3)
  gyroZInt       = DataSubsetDescriptor16B(FIFO_DATA, 5)
  tempInt        = DataSubsetDescriptor8B(FIFO_DATA, 7)

  gyroX          = CalculatedDescriptor(gyroXInt, getGyroValue, 'g')
  gyroY          = CalculatedDescriptor(gyroYInt, getGyroValue, 'g')
  gyroZ          = CalculatedDescriptor(gyroZInt, getGyroValue, 'g')
  temp           = CalculatedDescriptor(tempInt, FifoPkgFmt1Descriptor.getFifoTempValue, '°C')


class FifoPkgFmt3Descriptor(FifoPkgFmt1Descriptor):

  length = 16

  def __init__(self, basereg):
    super().__init__(basereg)
    self._fifoDataProp = getattr(FifoPkgFmt3Descriptor, 'FIFO_DATA')

  maskContF3 = FifoPkgFmt1Descriptor.maskAccel | FifoPkgFmt1Descriptor.maskGyro

  def parseData(self, obj, blob):
    header = blob[0]
    if header & self.maskContF3 != self.maskContF3:
      print('Invalid FIFO fmt 3 pkt')
      return None

    gaInts         = struct.unpack('>' + 'h'*6, struct.pack('B'*12, *blob[1:13]))

    accel          = FifoPkgFmt1Descriptor.getAccelValues(obj, gaInts[0:3])

    gyro           = FifoPkgFmt2Descriptor.getGyroValues(obj, gaInts[3:6])

    temp           = FifoPkgFmt1Descriptor.getFifoTempValue(blob[13])
    tmstInt        = struct.unpack('>H', struct.pack('BB', *blob[14:16]))[0]
    tmst           = self.unwrFifoTmstValue(None, tmstInt) * 1e-3 * 32 / 30

    data = {'values': accel + gyro + [ temp, tmst ] }
    return data

  def parseData2(self, obj, blob):
    data = self._fifoDataProp.splitr(obj, blob)
    if data['nomsg']:
      print('Invalid read: FIFO is empty')
      return None
    if not data['accel']:
      print('Configuration error: FIFO packet contains no accel data')
      return None
    if not data['gyro']:
      print('Configuration error: FIFO packet contains no gyro data')
      return None
    if data['is20b']:
      print('Configuration error: FIFO packet contains 20B data')
      return None
    if data['tmstFsync'] != 2:
      print('Configuration error: FIFO packet does not contain tmst')
      return None
    data['values'] = [ data[v] for v in ['accelX', 'accelY', 'accelZ', 'gyroX', 'gyroY', 'gyroZ', 'temp', 'tmst' ] ]
    return data

  def getTmstValue(_obj, value):
    return 1e-3 * 32.0 * value / 30 # ms

  def tstr(self):
    return 'Fifo Pkg Fmt III'

  FIFO_DATA      = RegMBDescriptor(0x30, length)
  header         = DataSubsetDescriptor8B(FIFO_DATA, 0)
  odrChGyro      = BitDescriptor(header, 0)
  odrChAccel     = BitDescriptor(header, 1)
  tmstFsync      = BitfieldDescriptor(header, 2, 2)
  is20b          = BitDescriptor(header, 4)
  gyro           = BitDescriptor(header, 5)
  accel          = BitDescriptor(header, 6)
  nomsg          = BitDescriptor(header, 7)

  accelXInt      = DataSubsetDescriptor16B(FIFO_DATA, 1)
  accelYInt      = DataSubsetDescriptor16B(FIFO_DATA, 3)
  accelZInt      = DataSubsetDescriptor16B(FIFO_DATA, 5)
  gyroXInt       = DataSubsetDescriptor16B(FIFO_DATA, 7)
  gyroYInt       = DataSubsetDescriptor16B(FIFO_DATA, 9)
  gyroZInt       = DataSubsetDescriptor16B(FIFO_DATA, 0xB)
  tempInt        = DataSubsetDescriptor8B(FIFO_DATA, 0xD)
  tmstInt        = DataSubsetDescriptor16B(FIFO_DATA, 0xE, dtype='H')

  accelX         = CalculatedDescriptor(accelXInt, FifoPkgFmt1Descriptor.getAccelValue, 'g')
  accelY         = CalculatedDescriptor(accelYInt, FifoPkgFmt1Descriptor.getAccelValue, 'g')
  accelZ         = CalculatedDescriptor(accelZInt, FifoPkgFmt1Descriptor.getAccelValue, 'g')
  gyroX          = CalculatedDescriptor(gyroXInt, FifoPkgFmt2Descriptor.getGyroValue, 'g')
  gyroY          = CalculatedDescriptor(gyroYInt, FifoPkgFmt2Descriptor.getGyroValue, 'g')
  gyroZ          = CalculatedDescriptor(gyroZInt, FifoPkgFmt2Descriptor.getGyroValue, 'g')

  temp           = CalculatedDescriptor(tempInt, FifoPkgFmt1Descriptor.getFifoTempValue, '°C')
  tmst           = CalculatedDescriptor(tmstInt, getTmstValue, 'ms')

  class UnwrFifoTmstValue:
    last = 0
    offs = 0
    def __call__(self, _, value):
      if value < self.last:
        self.offs += 2**16
      res = value + self.offs
      self.last = value
      return res
  unwrFifoTmstValue = UnwrFifoTmstValue()

class FifoPkgFmt4Descriptor(FifoPkgFmt1Descriptor):

  length = 20

  def __init__(self, basereg):
    super().__init__(basereg)
    self._fifoDataProp = getattr(FifoPkgFmt4Descriptor, 'FIFO_DATA')

  def parseData2(self, obj, blob):
    data = self._fifoDataProp.splitr(obj, blob)
    if data['nomsg']:
      print('Invalid read: FIFO is empty')
      return None
    if not data['accel']:
      print('Configuration error: FIFO packet contains no accel data')
      return None
    if not data['gyro']:
      print('Configuration error: FIFO packet contains no gyro data')
      return None
    if not data['is20b']:
      print('Configuration error: FIFO packet contain no 20B data')
      return None
    if data['tmstFsync'] != 2:
      print('Configuration error: FIFO packet does not contain tmst')
      return None
    data['values'] = [ data[v] for v in ['accelX', 'accelY', 'accelZ', 'gyroX', 'gyroY', 'gyroZ', 'temp', 'tmst' ] ]
    return data

  maskContF4 = FifoPkgFmt1Descriptor.mask20b | FifoPkgFmt1Descriptor.maskGyro | FifoPkgFmt1Descriptor.maskAccel
  maskExtL   = 0b111
  maskExtH   = 0b111000

  def parseData(self, obj, blob):
    header = blob[0]
    if header & self.maskContF4 != self.maskContF4:
      print('Invalid FIFO pkt')
      return None

    gaBases        = struct.unpack('>' + 'h'*6, struct.pack('B'*12, *blob[1:13]))

    tempInt        = struct.unpack('>h', struct.pack('BB', *blob[13:15]))[0]
    tmstInt        = struct.unpack('>H', struct.pack('BB', *blob[15:17]))[0]

    accelInt       = [ gaBases[i] << 3 | blob[17+i] & self.maskExtL for i in (0,1,2) ]
    gyroInt        = [ gaBases[3+i] << 3 | (blob[17+1] & self.maskExtH) >> 3 for i in (0,1,2) ]

    accel          = [ accelInt[i]  * self.accScale * obj._accelFactors[i] for i in obj._axes ]
    gyro           = [ gyroInt[i]  * self.gyroScale * obj._gyroFactors[i] for i in obj._axes ]

    temp           = self.getFifoTempValue(tempInt)
    tmst           = self.unwrFifoTmstValue(None, tmstInt) * 1e-3 * 32 / 30

    data = {'values': accel + gyro + [ temp, tmst ] }
    return data

  def tstr(self):
    return 'Fifo Pkg Fmt IV'

  FIFO_DATA      = RegMBDescriptor(0x30, length)
  header         = DataSubsetDescriptor8B(FIFO_DATA, 0)
  odrChGyro      = BitDescriptor(header, 0)
  odrChAccel     = BitDescriptor(header, 1)
  tmstFsync      = BitfieldDescriptor(header, 2, 2)
  is20b          = BitDescriptor(header, 4)
  gyro           = BitDescriptor(header, 5)
  accel          = BitDescriptor(header, 6)
  nomsg          = BitDescriptor(header, 7)

  accelXBase     = DataSubsetDescriptor16B(FIFO_DATA, 1)
  accelYBase     = DataSubsetDescriptor16B(FIFO_DATA, 3)
  accelZBase     = DataSubsetDescriptor16B(FIFO_DATA, 5)
  gyroXBase      = DataSubsetDescriptor16B(FIFO_DATA, 7)
  gyroYBase      = DataSubsetDescriptor16B(FIFO_DATA, 9)
  gyroZBase      = DataSubsetDescriptor16B(FIFO_DATA, 0xB)
  tempInt        = DataSubsetDescriptor16B(FIFO_DATA, 0xD)
  tmstInt        = DataSubsetDescriptor16B(FIFO_DATA, 0xF, dtype='H')
  extX           = DataSubsetDescriptor8B(FIFO_DATA, 0x11)
  extY           = DataSubsetDescriptor8B(FIFO_DATA, 0x12)
  extZ           = DataSubsetDescriptor8B(FIFO_DATA, 0x13)
  accelXExt      = BitfieldDescriptor(extX, 3, 0)
  gyroXExt       = BitfieldDescriptor(extX, 3, 3)
  accelYExt      = BitfieldDescriptor(extY, 3, 0)
  gyroYExt       = BitfieldDescriptor(extY, 3, 3)
  accelZExt      = BitfieldDescriptor(extZ, 3, 0)
  gyroZExt       = BitfieldDescriptor(extZ, 3, 3)

  accScale       = 16/(2**18)
  accelXInt      = JointBitfieldDescriptor([accelXBase, accelXExt])
  accelYInt      = JointBitfieldDescriptor([accelYBase, accelYExt])
  accelZInt      = JointBitfieldDescriptor([accelZBase, accelZExt])
  accelX         = ScaledDescriptor(accelXInt, accScale, 'g')
  accelY         = ScaledDescriptor(accelYInt, accScale, 'g')
  accelZ         = ScaledDescriptor(accelZInt, accScale, 'g')

  gyroScale      = 2000/(2**18)
  gyroXInt       = JointBitfieldDescriptor([gyroXBase, gyroXExt])
  gyroYInt       = JointBitfieldDescriptor([gyroYBase, gyroYExt])
  gyroZInt       = JointBitfieldDescriptor([gyroZBase, gyroZExt])
  gyroX          = ScaledDescriptor(gyroXInt, gyroScale, 'g')
  gyroY          = ScaledDescriptor(gyroYInt, gyroScale, 'g')
  gyroZ          = ScaledDescriptor(gyroZInt, gyroScale, 'g')

  def getFifoTempValue(self, value):
    return value / 132.48 + 25
  temp           = CalculatedDescriptor(tempInt, getFifoTempValue, '°C')

  unwrFifoTmstValue= FifoPkgFmt3Descriptor.UnwrFifoTmstValue()

  tmstIntUnwr    = CalculatedDescriptor(tmstInt, FifoPkgFmt3Descriptor.UnwrFifoTmstValue(), '')
  tmst           = ScaledDescriptor(tmstIntUnwr, 1e-3 * 32 / 30, 'ms')
