import argparse
import sys
import struct
import time
import socket
import fcntl
import termios
import signal

from .spitool import SPIConn
from .iim42652 import IIM42652

from . import linesocket
from . import log
from . import queue

from .timer import WTimer

maxReadBlkSize = 1024 + 512
blksz = 12

class Status:
  done = False
  stopped = False
  fcnt = wmfl = ovfl = 0
  maxfcnt = avgfcnt = stdfcnt = nfifodata = 0
  nsamples = nreads = nresets = nwmfl = 0
  curTempVals = [0]

status = Status()
timer = WTimer()

def getFIFO(iim42652):
  fcnt, wmfl, ovfl = iim42652.getFIFOInfo()
  #  print('FIFO status', fcnt, wmfl, ovfl)
  rcnt = int(min(fcnt,maxReadBlkSize)/blksz)
  if rcnt > 0:
    v = iim42652.getFIFOPacketData(rcnt)
    fifodata = struct.pack('B'*rcnt*blksz, *v)
    status.fcnt = fcnt
    status.maxfcnt = max(fcnt, status.maxfcnt)
    # assert(len(fifodata) == rcnt)
    # assert(len(v) == rcnt)
    # assert(rcnt % blksz == 0)
  else:
    fifodata = b''

  if ovfl:
    iim42652.resetFIFO()
    log.error("FIFO overflow: FIFO %d" % (fcnt,))
    status.nresets += 1
    status.maxfcnt = 0
    sys.stdout.flush()
  elif wmfl:
    status.nwmfl += 1
#      log.error("FIFO watermark reached: it %d/%d, FIFO %d, buf %d" % (iloop, fetch, fcnt, len(fifodata)))

  return fifodata


def readFIFO(sensor, ls):
  global blksz
  timer.reset()

  blksz = sensor.getFIFOBlockSize()

  log.msg('blksz', blksz)

#  print(dir(socket))
#  print(dir(termios))

#  if sensor.tempmanrfreq > 0:
#    tempManInter = 1 / sensor.tempmanrfreq
#    status.curTempVals = [sensor.readTemp(),]

  smplrt = sensor.fifoSampleRate
  smpldt = 1 / smplrt

  que = queue.Queue(1e6)

  sensor['fifoMode'] = 'Stream-to-FIFO'

  while not status.done or not que.empty():
    if status.done:
      if not status.stopped:
        log.msg('Stopping queue')
        sensor['fifoMode'] = 'bypass'
        fifodata = b''
        status.stopped = True
    else:
      fifodata = getFIFO(sensor)
      status.nfifodata += len(fifodata)

    timer.click()

    if len(fifodata) > 0:
      try:
        que.put(fifodata)
      except:
        pass
    elif que.empty():
      time.sleep(4e-4)

    for i in range(min(3,que.qsize())):
      try:
        item = que.peek()
        ls.transmitb(item)
        que.get()
      except BlockingIOError as ex:
        #print('socket blocked for queue')
        break
      except BrokenPipeError as ex:
        print('socket closed, exit')
        status.done = True
        break

    if len(fifodata) < 1000:
      if smplrt < 100:
        time.sleep(smpldt/2)
      else:
        time.sleep(smpldt * 10)

#    if status.fcnt < 1000 and sensor.tempmanrfreq > 0 and timer.elapsed('tempfetch', tempManInter):
#      status.curTempVals = [sensor.readTemp(),]

    if timer.elapsed('info', 1) or timer.last > 20:
      nwrite = 0
      # nwrite = ls.sock.getsockopt(socket.SOL_SOCKET, socket.SO_NWRITE, 8)
      print("%s(% 3d)%s: %d B, %.2f KB/s, FIFO %d (max %d), ov %d wm %d, queue %d" %
            (">" if 1e-3 * timer.last > smpldt * 40 else " ", 1e-3 * timer.last / smpldt, timer,
             status.nfifodata, 1e-3 * status.nfifodata / timer.elaps,
             status.fcnt, status.maxfcnt, status.nresets, status.nwmfl, que.qsize()))
      sys.stdout.flush()

def ctrlCHandler(sig, frame):
  log.msg(f'Signal {sig} caught, setting stop')
  status.done = True
  return

def run():
  import os
  log.name = 'fiforeader'

  if os.geteuid() == 0:
    res = os.setpriority(os.PRIO_PROCESS, 0, -20)
    print('Am root', res)

  parser = argparse.ArgumentParser(description='FIFO reader helper process')

  parser.add_argument('--bus', dest='bus', type=int, default=0, help='SPI bus')
  parser.add_argument('--dev', dest='dev', type=int, default=0, help='SPI device')
  parser.add_argument('--speed', dest='speed', type=int, default=int(1e6), help='SPI speed')

  parser.add_argument('--outfile', type=str, help='name of output file')

  parser.add_argument('--fifo-packet', type=int)

  args = parser.parse_args(sys.argv[1:])
  print(args)

  signal.signal(signal.SIGINT, ctrlCHandler)
  signal.signal(signal.SIGTERM, ctrlCHandler)

  spi = SPIConn(args.bus, args.dev, args.speed)
  print(spi)

  iim42652 = IIM42652(spi)
  print(iim42652)

  iim42652.setupFIFO(args.fifo_packet, iim42652.sampleRate)

  if not iim42652.connected:
    log.error("iim42652 device not found")
    return
  else:
    log.msg("iim42652 device connected")

  ls = linesocket.LineSocket(args.outfile)
  ls.openSocket(1)

  try:
    readFIFO(iim42652, ls)
  except KeyboardInterrupt as ex:
    log.msg('FifoReader caught Ctrl-C')
  except BaseException as ex:
    log.msg(f'FifoReader caught exception {ex}')

  log.msg(f'FifoReader read {status.nfifodata} B, {status.nfifodata/blksz} samples from FIFO')
  log.msg('FifoReader exiting')

if __name__ == "__main__":
    run()
