#! /usr/bin/env python3

from . import spitool
import time
import datetime
import argparse
import math
import re

from iim42652_spi.iim42652 import IIM42652

def parseargs():
  parser = argparse.ArgumentParser(description='Erzeuge Umspul-Konfiguration.')

  parser.add_argument('--bus', dest='bus', type=int, default=0, help='SPI bus')
  parser.add_argument('--dev', dest='dev', type=int, default=0, help='SPI device')
  parser.add_argument('--speed', dest='speed', type=int, default=int(1e6), help='SPI speed')

  parser.add_argument('--trace', action='store_true', help='Trace SPI')
  parser.add_argument('--reset', action='store_true', help='Reform initial reset')

  parser.add_argument('--odr', '-r', type=int, default=100, help='Trace SPI')

  parser.add_argument('--wom', action='store_true', help='Trace SPI')
  parser.add_argument('--smd-short', action='store_true', help='Trace SPI')
  parser.add_argument('--smd-long', action='store_true', help='Trace SPI')

  parser.add_argument('--axes', '-a', type=int, nargs='+', help='Trace SPI')

  parser.add_argument('--mode', type=str, help='Trace SPI')
  parser.add_argument('--int-mode', type=str, help='Trace SPI')

  parser.add_argument('--any', action='store_true', help='Trace SPI')
  parser.add_argument('--all', action='store_true', help='Trace SPI')

  args = parser.parse_args()
  return args

def run():
  import locale
  locale.setlocale(locale.LC_ALL, '')

  args = parseargs()
  if args.trace:
    print(args)
    print(f'locale: {locale.getlocale()}')
    IIM42652.DEBUG = True
    spitool.DEBUG = True

  spi = spitool.SPIConn(args.bus, args.dev, args.speed)

  iim42652 = IIM42652(spi)
  if not iim42652.connected:
    print("device not found")
    return

  if args.reset:
    iim42652.reset()

  odr = 50
  if args.odr:
    odr = args.odr
  iim42652['accelODR'] = odr
  iim42652['powerAccel'] = 'lowpower'
  iim42652['dmpODR'] = odr
  iim42652['dmpPowerSave'] = False
  iim42652['pedEnable'] = True
  time.sleep(1e-3)

  iim42652['a_disable'] = False

  iim42652['dmpMemResetEn'] = True
  time.sleep(1e-3)

  # defaults as per the docs
  iim42652['lowEnergyAmpThreshSel'] = 10
  iim42652['pedAmplThreshSel'] = 8
  iim42652['pedStepCntThreshSel'] = 5
  iim42652['pedHiEnThreshSel'] = 1
  iim42652['pedSBTimerThreshSel'] = 4
  iim42652['pedStepDetThreshSel'] = 2
  iim42652['sensitivityModeInt'] = 0
  
  iim42652['dmpInitEn'] = True
  time.sleep(50e-3)

  iim42652['INT_SOURCE6'] = 1 << 5

  iim42652['ffEnable'] = False
  iim42652['pedEnable'] = True

  inter = 0.5 / odr

  nInter = 0
  lastSteps = 0

  print(f'pedometer counting steps')
  while True:
    INT3 = iim42652['INT_STATUS3']
    stepInt = INT3 & (1 << 5)
    stepCntOvfInt = INT3 & (1 << 4)
    if stepCntOvfInt:
      nInter += 1
    if not stepInt:
      time.sleep(inter)
      continue
    t1 = datetime.datetime.now()
    steps = iim42652['stepCountFFDur'] + nInter * (2**16 - 1)
    stepCad = iim42652['stepCadence']
    cadHz = 0 if stepCad == 0 else odr / stepCad
    actClass = iim42652['activityClass']
    if steps != lastSteps:
      print(f'{t1.strftime("%x %X.%f")} steps: {steps} cadence: {cadHz} Hz activity: {actClass}')
      lastSteps = steps
    time.sleep(inter)

if __name__ == "__main__":
  run()
