#! /usr/bin/env python3

import time
import math
import argparse

from . import spitool
from . import table
from . import timer
from .iim42652 import IIM42652
from .propcache import PropCache

def parseargs():
  parser = argparse.ArgumentParser(description='Erzeuge Umspul-Konfiguration.')

  parser.add_argument('--bus', dest='bus', type=int, default=0, help='SPI bus')
  parser.add_argument('--dev', dest='dev', type=int, default=0, help='SPI device')
  parser.add_argument('--speed', dest='speed', type=int, default=int(1e6), help='SPI speed')

  parser.add_argument('--trace', action='store_true', help='Trace SPI')

  parser.add_argument('--outfile', '-o', type=str, default="out.csv", help='Trace SPI')
  parser.add_argument('--precision', '-p', type=int, help='CSV precision')
  parser.add_argument('--format', type=str, help='CSV value format (fixed or adaptive)')

  parser.add_argument('--maxreads', type=int, help='number of reads')
  parser.add_argument('--maxtime', type=float, help='time limit of reading')

  parser.add_argument('--odr', '-r', type=int, help='Trace SPI')
  parser.add_argument('--scale', '-s', type=int, help='Trace SPI')

  defaults = {
    'outfile': 'out.csv',
    'precision': 8,
    'format': 'eng',
    'maxreads': -1,
    'maxtime': -1,
    'odr': 2000,
    'scale': 16
  }

  parser.set_defaults(**defaults)

  args = parser.parse_args()
  return args



def run():
  import locale
  locale.setlocale(locale.LC_ALL, '')

  args = parseargs()
  if args.trace:
    print(args)
    print(f'locale: {locale.getlocale()}')
    IIM42652.DEBUG = True
    spitool.DEBUG = True

  spi = spitool.SPIConn(args.bus, args.dev, args.speed)

  iim42652 = IIM42652(spi)
  if not iim42652.connected:
    print("device not found")
    return

  i = 0
  i0 = ilast = 0
  t0 = tlast = time.time()

  header = ('Time (s)', 'Acc_x (g)', 'Acc_y (g)', 'Acc_z (g)')
  outf = open(args.outfile, 'w')
  outt = table.Table(outf, header, args.precision)
  outt.writeHeader()

  tout = timer.WTimer()
  tout.click()
  tlastp = tout.time

  aProp = getattr(IIM42652.Register, 'ACCEL_3D_DATA')

  iim42652['powerGyro'] = 'off'
  iim42652['powerAccel'] = 'lownoise'
  iim42652['accelODR'] = args.odr
  odr = iim42652['accelODR']

  print(f'ODR: setting to {args.odr} Hz, set to {odr} Hz')

  time.sleep(10e-3)
  tout.click()
  t0 = tout.time

  cache = PropCache(iim42652.register)

  while True:
    tout.click()

    iim42652.wait4data()

    ablob = iim42652['ACCEL_3D_DATA']
    adata = aProp.splitr(iim42652.register, ablob)
    data = tuple(adata[c] for c in ('accelDataX', 'accelDataY', 'accelDataZ'))

    timeval = tout.elaps
    data = (timeval,) + data

    outt.writeValues(data)

    if tout.elapsed('info', 1):
      cnt = i - ilast
      datastr = str([round(v*100)/100 for v in data])
      print(f"{i: 7d}: {datastr:>30s} {cnt / (tout.time - tlastp):.2f} reads/s {i / (tout.time - t0):.2f} reads/s f.s.")
      tlastp = tout.time
      ilast = i
#    else:
#      time.sleep(0.5 / odr)

    i += 1

    if args.maxreads > 0 and i > args.maxreads:
      break
    if args.maxtime >= 0 and tout.elaps > args.maxtime:
      break

if __name__ == "__main__":
  run()

# Local Variables:
# coding: utf-8
# End:
