#! /usr/bin/env python3

from . import spitool
import time
import datetime
import argparse
import math
import re

from iim42652_spi.iim42652 import IIM42652

def parseargs():
  parser = argparse.ArgumentParser(description='Erzeuge Umspul-Konfiguration.')

  parser.add_argument('--bus', dest='bus', type=int, default=0, help='SPI bus')
  parser.add_argument('--dev', dest='dev', type=int, default=0, help='SPI device')
  parser.add_argument('--speed', dest='speed', type=int, default=int(1e6), help='SPI speed')

  parser.add_argument('--trace', action='store_true', help='Trace SPI')
  parser.add_argument('--reset', action='store_true', help='Reform initial reset')

  parser.add_argument('--odr', '-r', type=int, help='Trace SPI')

  parser.add_argument('--wom', action='store_true', help='Trace SPI')
  parser.add_argument('--smd-short', action='store_true', help='Trace SPI')
  parser.add_argument('--smd-long', action='store_true', help='Trace SPI')

  parser.add_argument('--threshold', '-t', type=float, nargs='+', help='Trace SPI')

  parser.add_argument('--axes', '-a', type=int, nargs='+', help='Trace SPI')

  parser.add_argument('--mode', type=str, help='Trace SPI')
  parser.add_argument('--int-mode', type=str, help='Trace SPI')

  parser.add_argument('--any', action='store_true', help='Trace SPI')
  parser.add_argument('--all', action='store_true', help='Trace SPI')

  args = parser.parse_args()
  return args

def run():
  import locale
  locale.setlocale(locale.LC_ALL, '')

  args = parseargs()
  if args.trace:
    print(args)
    print(f'locale: {locale.getlocale()}')
    IIM42652.DEBUG = True
    spitool.DEBUG = True

  spi = spitool.SPIConn(args.bus, args.dev, args.speed)

  iim42652 = IIM42652(spi)
  if not iim42652.connected:
    print("device not found")
    return

  if args.reset:
    iim42652.reset()

  iim42652['powerAccel'] = 'lownoise'
  iim42652['smdMode'] = 'off'
  iim42652['dmpInitEn'] = True
  iim42652['dmpMemResetEn'] = True
  dummy = iim42652['womInt'] # clear interrupts
  dummy = iim42652['smdInt'] # clear interrupts

  mode = 'wom'
  if args.wom:
    mode = 'wom'
  elif args.smd_short:
    mode = 'smdShort'
  elif args.smd_long:
    mode = 'smdLong'

  womMode = 1
  if args.mode:
    if 'init' in args.mode:
      womMode = 0

  intmode = 0 # or
  if args.int_mode:
    if 'all' in args.int_mode:
      intmode = 1

  axes = range(3)
  if args.axes:
    axes = args.axes
  axDisable = [False] * 3
  for i in range(3):
    if i not in axes:
      axDisable[i] = True
  iim42652['a_disable'] = axDisable

  odr = 50
  if args.odr:
    odr = args.odr
  iim42652['accelODR'] = odr
  iim42652['dmpODR'] = odr
  iim42652['powerAccel'] = 'lowpower'
  time.sleep(1e-3)

  threshold = [0.1]
  if args.threshold:
    threshold = args.threshold
  if len(threshold) == 1:
    iim42652['accelWomThr'] = threshold[0]
  elif len(threshold) == 3:
    iim42652['accelWomThr'] = threshold
  time.sleep(1e-3)

  time.sleep(50e-3)
  iim42652['smdMode'] = mode
  iim42652['womModeInt'] = womMode
  iim42652['womIntModeInt'] = intmode

  print(f'waiting for {mode} events')
  if mode == 'wom':
    while True:
      womInts = iim42652['womInt']
      if (intmode == 0 and any(womInts)) or (intmode == 1 and all(womInts)):
        t1 = datetime.datetime.now()
        print(f'{t1.strftime("%x %X")} WOM event: {womInts}')
      time.sleep(1 / odr)
  else:
    while True:
      if iim42652['smdInt']:
        womInts = iim42652['womInt']
        t1 = datetime.datetime.now()
        print(f'{t1.strftime("%x %X")} SMD event')
      time.sleep(1 / odr)

if __name__ == "__main__":
  run()
