#! /usr/bin/env python3

from . import spitool
import time
import argparse
import math
import re

from iim42652_spi.iim42652 import IIM42652

def parseargs():
  parser = argparse.ArgumentParser(description='Erzeuge Umspul-Konfiguration.')

  parser.add_argument('--bus', dest='bus', type=int, default=0, help='SPI bus')
  parser.add_argument('--dev', dest='dev', type=int, default=0, help='SPI device')
  parser.add_argument('--speed', dest='speed', type=int, default=int(1e6), help='SPI speed')

  parser.add_argument('--trace', action='store_true', help='Trace SPI')
  parser.add_argument('--reset', action='store_true', help='Perform initial reset')

  parser.add_argument('--test', action='store_true', help='Perform self test')
  parser.add_argument('--connected', action='store_true', help='Check if device connected')

  parser.add_argument('--show', '-s', action='store_true', help='Show SPI register and property info')
  parser.add_argument('--dump', '-d', action='store_true', help='Dump properties')
  parser.add_argument('--dump-all', '-D', action='store_true', help='Dump all SPI registers and properties')
  parser.add_argument('--register', '-r', '--dump-register', action='store_true', help='Dump all SPI registers')

  parser.add_argument('--bank', '-b', type=int, nargs='+', help='Dump only fields from selected banks')
  parser.add_argument('--filter', '-f', type=str, nargs='+', help='Dump only fields whose names match filter')
  parser.add_argument('--ignorecase', '-i', action='store_true', help='Filters ignore case')

  parser.add_argument('property', metavar='prop', type=str, nargs='*', help='Property or bank:reg')

  args = parser.parse_args()
  return args

def printPropVal(iim42652, prop, val, indent=''):
  name = f'{indent}{prop.name}'
  if hasattr(prop, 'width'):
    hexStrFmt = '{val:0%dX}' % (2 if prop.width > 4 else 1)
    hexStr    = hexStrFmt.format(val=val)
    binStrFmt = '{val:0%db}' % (prop.width)
    binStr    = binStrFmt.format(val=val)
    print(f'{name:28s} = {val} (0x{hexStr}, 0b{binStr})')
  else:
    if type(val) == type(True):
      val = int(val)
    print(f'{name:28s} = {val} {prop.unit}')

def handleProperty(iim42652, args, propName):
  val = None
  if '=' in propName:
    (propName, val) = (f.strip() for f in propName.split('='))
  elif ' ' in propName:
    (propName, val) = (f for f in propName.split(' ') if f != '')
  if propName in iim42652:
    prop = getattr(IIM42652.Register, propName)
    if args.show:
      print(prop)
    if val:
      try:
        if val.startswith('['):
          val = eval(val)
        elif val.startswith('0x'):
          val = int(val[2:-1], 16)
        elif any([ val.startswith(digit) for digit in '0123456789-' ]):
          if '.' in val or 'e' in val:
            val = float(val)
          else:
            val = int(val)
      except BaseException as ex:
        print(ex)
        val = True if val == 'True' else False if val == 'False' else val
      setattr(iim42652.register, propName, val)
    val = getattr(iim42652.register, propName)
    printPropVal(iim42652, prop, val)
  else:
    print(f'property {propName} not found')
    exit(1)

namew = 20
def printFieldTree(iim42652, args, field, dumpAll=False, noRec=False, reFilts=[], indent=''):
  prop = getattr(IIM42652.Register, field)
  if not dumpAll:
    indent = ''
  if (len(prop.children) == 0 or dumpAll or noRec or len(reFilts)) \
     and (len(reFilts) == 0 or any([ f.search(field) for f in reFilts ])):
    val = iim42652[field]
    if args.show:
      print(f'{indent}{prop}')
    printPropVal(iim42652, prop, val, indent)
  if noRec:
    return
  for c in prop.children:
    if prop.children[c] is not None:
      printFieldTree(iim42652, args, c, dumpAll=dumpAll, noRec=noRec, reFilts=reFilts, indent=indent + ' ')

def run():
  args = parseargs()
  if args.trace:
    print(args)
    IIM42652.DEBUG = True
    spitool.DEBUG = True

  spi = spitool.SPIConn(args.bus, args.dev, args.speed)

  iim42652 = IIM42652(spi)
  if not iim42652.connected:
    print("device not found")
    exit(1)

  if args.reset:
    iim42652.reset()

  if args.show and args.property is None:
    print(iim42652)
    return

  if args.test:
    print(f'Performing self tests...')
    ra = iim42652.accelSelfTest()
    if not ra[0]:
      print(f'Accel self test failed, relative errors > 5%: {ra[3]}')
    rg = iim42652.accelSelfTest()
    if not rg[0]:
      print(f'Gyro self test failed, relative errors > 5%: {rg[3]}')
    if not (ra[0] and rg[0]):
      exit(1)

  for prop in args.property:
    handleProperty(iim42652, args, prop)

  banks = range(5) if args.bank is None else args.bank
  reFilts = []
  if args.filter:
    reFilts = [re.compile(filter, re.IGNORECASE if args.ignorecase else 0) for filter in args.filter]

  if args.dump or args.dump_all or args.register or args.filter:
    iim42652.register.USER_BANK = 0
    for field in iim42652:
      if field[0] in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ' \
         and getattr(IIM42652.Register, field).bank in banks:
        printFieldTree(iim42652, args, field, args.dump_all, args.register, reFilts)

if __name__ == "__main__":
  run()
