# author: Johannes Willkomm <j.willkomm@fionec.de>

import json

def flatten(data):
    if isinstance(data, list):
        if len(data) == 0:
            return ()
        else:
            return flatten(data[0]) + flatten(data[1:])
    elif isinstance(data, tuple):
        if len(data) == 0:
            return ()
        else:
            return flatten(data[0]) + flatten(data[1:])
    else:
        return (data,)

serFields = ['headers', 'colWidths', 'colPrecs', 'valfmts']

class Table:

    colWidths = [8, 8, 8]
    sep = ';'
    _nchar = 0

    def __init__(self, fd, headers, colPrec, sep=';', format='var'):
        self.fd = fd
        self.headers = headers

        self.sep = sep
        self.format = format

        self.fmtletter = 'f' if format == 'fixed' else 'e' if format == 'eng' else 'g'
        self.fmtaddspace = 3 if format == 'fixed' else 7 if format == 'eng' else 5

        if type(colPrec) == type([]) or type(colPrec) == type(()):
            self.colPrecs = colPrec
        else:
            self.colPrecs = [colPrec] * len(headers)

        self.colWidths = [v + self.fmtaddspace for v in self.colPrecs]

        self.valfmts = ['%%.%d%s' % (p, self.fmtletter) for p in self.colPrecs]

        self.update()

    def update(self):
        self.lnFormat = self.sep.join(['%% %ds' % w for w in self.colWidths])

    def writeHeader(self):
        self.write('%s\n' % self.getHeader())
        self.fd.flush()

    def getHeader(self):
        return self.lnFormat % self.headers

    def writeLine(self, vline):
        self.write('%s\n' % vline)

    def writeSamples(self, values):
        vline = self.lnFormat % values
        self.writeLine(vline)

    def writeValues(self, values):
        self.writeLine(self.fmtValues(values))

    def fmtValues(self, values):
        return self.lnFormat % tuple((self.valfmts[i] % f) for (i,f) in enumerate(values))

    def write(self, res):
        self._nchar += len(res)
        self.fd.write(res)

    @property
    def nchar(self):
        return self._nchar

    def __str__(self):
        s = ''
        s += 'sep: ' + self.sep + '\n'
        s += 'headers (%d): ' % (len(self.headers),) + str(self.headers) + '\n'
        s += 'colPrecs (%d): ' % (len(self.colPrecs),) + str(self.colPrecs) + '\n'
        s += 'colWidths (%d): ' % (len(self.colWidths),) + str(self.colWidths) + '\n'
        s += 'valfmts (%d): ' % (len(self.valfmts),) + str(self.valfmts) + '\n'
        s += 'lnFormat: ' + self.lnFormat + '\n'
        return s

    def __repr__(self):
        tdata = {v: getattr(self, v) for v in serFields}
        return json.dumps(tdata, indent=1)

    def readJSON(self, jsdata):
        tdata = json.loads(jsdata)
        for i in tdata:
            if hasattr(self, i):
                setattr(self, i, flatten(tdata[i]))
        self.update()

    def getV3Table(self, newfd):
        v3Table = Table(newfd, self.headers, self.colPrecs, self.sep, self.format)

        v3Table.headers = flatten([(h + ' (min)', h + ' (mean)', h + ' (max)') for h in self.headers])
        v3Table.colPrecs = flatten([(c,c,c) for c in self.colPrecs])
        v3Table.colWidths = flatten([(c,c,c) for c in self.colWidths])
        v3Table.valfmts = flatten([(c,c,c) for c in self.valfmts])

        v3Table.update()

        return v3Table


if __name__ == "__main__":
    import sys
    t = Table(sys.stdout, [12, 4, 5, 7, 12])
    t.writeHeader(('A', 'B', 'C', 'D', 'E'))
    t.writeLine(('1', '2', '3', '4', '5'))
    t.writeLine(('1', '2', '3', '4', '5'))
